/*
 * midimsg.c -- Midi event parsing code
 *
 * $Id: midievents.c,v 1.5 1998/02/06 09:49:31 johns Exp $
 *
 */

#include <stdio.h>
#include <string.h>
#include <glib.h>

#include "midi.h"
#include "midievents.h" 

MidiHandle *
midi_handle_new(void (*handler) (MidiHandle *))
{
  MidiHandle * handle;

  handle = g_new0 (MidiHandle, 1);

  handle->handler = handler;
 
  return handle;
} 

void
midi_handle_free (MidiHandle * handle) {
  g_free (handle);
} 

int
midi_get_status (MidiHandle *handle)
{
  g_return_val_if_fail (handle->buflen > 0, -1);
  return (handle->buf[0] & MIDI_STATUS_TYPE_MASK);
}

int
midi_get_channel (MidiHandle * handle) 
{
  g_return_val_if_fail (handle->buflen > 0, -1);
  return (handle->buf[0] & MIDI_CHAN_MASK) + 1;
}

int
midi_get_note (MidiHandle * handle)
{
  g_return_val_if_fail (handle->buflen > 0, -1);
  /* TODO: assert right type of MIDI event */

  return (handle->buf[1] & MIDI_DATA_MASK);
}

int
midi_get_velocity (MidiHandle *handle) 
{
  g_return_val_if_fail (handle->buflen > 0, -1);
  /* TODO: assert right type of MIDI event */

  return (handle->buf[2] & MIDI_DATA_MASK);
}

int
midi_get_controller (MidiHandle *handle)
{
  g_return_val_if_fail (handle->buflen > 0, -1);
  /* TODO: assert right type of MIDI event */

  return (handle->buf[1] & MIDI_DATA_MASK);
}

int
midi_get_value (MidiHandle *handle)
{
  g_return_val_if_fail (handle->buflen > 0, -1);

  return (handle->buf[2] & MIDI_DATA_MASK);
}

int 
midi_get_program (MidiHandle *handle) 
{
  g_return_val_if_fail (handle->buflen > 0, -1);
  /* TODO: assert right type of MIDI event */

  return (handle->buf[1] & MIDI_DATA_MASK);
}

int
midi_get_chan_pressure (MidiHandle *handle) 
{
  g_return_val_if_fail (handle->buflen > 0, -1);
  /* TODO: assert right type of MIDI event */

  return (handle->buf[1] & MIDI_DATA_MASK);
}

int
midi_get_pitch_bend (MidiHandle *handle)
{
  guint tmp;

  g_return_val_if_fail (handle->buflen > 1, -1);

  tmp = handle->buf[1];
  tmp |= (handle->buf[2] << 7);
  
  return tmp;
}

void midi_default_handler(MidiHandle * handle)
{
  gint velocity;

  switch (midi_get_status (handle)) 
    {
    case MIDI_NOTE_OFF:
      printf("Note off: ");
      printf("channel %d ", midi_get_channel (handle));
      printf("note %d ", midi_get_note (handle));
      printf("velocity %d ", midi_get_velocity (handle));
      printf("\n");
      break;

    case MIDI_NOTE_ON:
      if ((velocity = midi_get_velocity (handle)) != 0) {
        printf("Note on: ");
        printf("channel %d ", midi_get_channel (handle));
        printf("note %d ", midi_get_note (handle));
        printf("velocity %d ", velocity);
        printf("\n");
      }
      else {
        printf("Note off: ");
        printf("channel %d ", midi_get_channel (handle));
        printf("note %d ", midi_get_note (handle));
        printf("velocity %d ", velocity);
        printf("\n");
      }
      break;

    case MIDI_KEY_PRESSURE:
      printf("Key Pressure: ");
      printf("channel %d ", midi_get_channel (handle));
      printf("note %d ", midi_get_note (handle));
      printf("pressure %d ", midi_get_velocity (handle));
      printf("\n");
      break;

    case MIDI_CTRL_CHANGE:
      printf("Control Change: ");
      printf("channel %d ", midi_get_channel (handle));
      printf("controller %d ", midi_get_controller (handle));
      printf("value %d ", midi_get_value (handle));
      printf("\n");
      break;

    case MIDI_PITCH_BEND:
      printf("Pitch Bend: ");
      printf("channel %d ", midi_get_channel (handle));
      printf("value %d ", midi_get_pitch_bend (handle)); 
      printf("\n");
      break;

    /* Messages with 1 data byte */
    case MIDI_PROG_CHANGE:
      printf("Program Change: ");
      printf("channel %d ", midi_get_channel (handle));
      printf("program %d ", midi_get_program (handle));
      printf("\n");
      break;

    case MIDI_CHAN_PRESSURE:
      printf("Channel Pressure: ");
      printf("channel %d ", midi_get_channel (handle));
      printf("pressure %d ", midi_get_chan_pressure (handle));
      printf("\n");
      break;

    /* Special MIDI messages, can be varying length */
    case MIDI_SYSEX: /* should dispatch handlers for this stuff */
      printf("System Exclusive\n");
      break;

    default:
      printf("Unknown MIDI Status\n");
      break;
  }
}

void 
midi_add_byte (MidiHandle *handle, guchar b)
{
  if (b & MIDI_STATUS_MASK)
    {
    switch (b & MIDI_STATUS_TYPE_MASK)
      {
      case MIDI_NOTE_OFF:
      case MIDI_NOTE_ON:
      case MIDI_KEY_PRESSURE:
      case MIDI_CTRL_CHANGE:
      case MIDI_PITCH_BEND:
        handle->buf[0] = b;
        handle->bufpos=1;
        handle->buflen=3; 
        break;

      /* Messages with 1 data byte */
      case MIDI_PROG_CHANGE:
      case MIDI_CHAN_PRESSURE:
        handle->buf[0] = b;
        handle->bufpos=1;
        handle->buflen=2; 
        break;

      /* Special MIDI messages, can be varying length */
      case MIDI_SYSEX: /* should dispatch handlers for this stuff */
if (b != 0xf8)
  printf("System Message: %2x\n", b); 
        handle->buf[0]=0;
        handle->bufpos=0;
        handle->buflen=0;
        break;
        
      /* Out of sync, got data without matching status */
      default: 
printf("Unknown MIDI Status Message\n");
        handle->buf[0]=0;
        handle->bufpos=0;
        handle->buflen=0;
        break;
    }

    return;
  } 

  /* got a MIDI data byte, but didn't get the status byte that went with it */
  if (handle->buf[0] == 0) 
    return; 

  if (handle->bufpos < handle->buflen) {
    handle->buf[handle->bufpos] = b; /* add data byte to buffer */
    handle->bufpos++;
  }

  if (handle->bufpos >= handle->buflen) {
    handle->handler(handle);		/* do something with the MIDI event */
    handle->bufpos=1;
  }
}


