/* keytonote.c - GTK+ "KeyToNote" Container Widget
 * Copyright (C) 1998 Simon Kgedal
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License in the file COPYING for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <gtk/gtksignal.h>
#include <gdk/gdkkeysyms.h>
#include "util.h"
#include "noteinput.h"
#include "glib_1_0_compat.h"

static GtkEventBoxClass *parent_class = NULL;

/* Signals */
enum
{
  NOTE_INPUT_PRESS,
  NOTE_INPUT_RELEASE,
  NOTE_INPUT_TRANSPOSE_CHANGE,
  LAST_SIGNAL
};

static gint note_input_signals[LAST_SIGNAL] = {0};

typedef void (*NoteInputSignal1) (GtkObject *object,
				  gint arg1,
				  gpointer data);

/* Signal Marshalers */
static void
note_input_marshal_signal_1 (GtkObject *object,
			     GtkSignalFunc func,
			     gpointer func_data,
			     GtkArg *args);

/* NoteInput Methods */
static void note_input_class_init (NoteInputClass *class);
static void note_input_init (NoteInput *noteinput);

/* GtkObject Methods */
static void note_input_destroy (GtkObject *object);

/* GtkWidget Methods */
static gint note_input_key_press (GtkWidget *widget, GdkEventKey *event);
static gint note_input_key_release (GtkWidget *widget, GdkEventKey *event);

/* Stuff that can happen when you press a key */
typedef enum
{
  NI_FUNC_KEY,			/* `data' is MIDI key to press/release */
  NI_FUNC_TRANSPOSE_UP,
  NI_FUNC_TRANSPOSE_DOWN,
  NI_FUNC_TRANSPOSE_UP_OCT,
  NI_FUNC_TRANSPOSE_DOWN_OCT,
  NI_FUNC_OCTAVE_N,		/* goto octave `data' */
  NI_FUNC_NONE			/* couldn't find any */
} NIFuncType;

typedef struct _NIFunc NIFunc;

struct _NIFunc
{
  NIFuncType type;
  gint data;
};

/***********************************************************************/

GtkType
note_input_get_type (void)
{
  static GtkType note_input_type = 0;

  if (!note_input_type)
    {
      GtkTypeInfo note_input_info = 
      {
	"NoteInput",
	sizeof (NoteInput),
	sizeof (NoteInputClass),
	(GtkClassInitFunc) note_input_class_init,
	(GtkObjectInitFunc) note_input_init,
	(GtkArgSetFunc) NULL,
	(GtkArgSetFunc) NULL,
      };

      note_input_type = 
	gtk_type_unique (gtk_event_box_get_type (), &note_input_info);
    }

  return note_input_type;
}

static void
note_input_class_init (NoteInputClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass *) class;
  widget_class = (GtkWidgetClass *) class;

  parent_class = gtk_type_class (gtk_event_box_get_type ());

  note_input_signals[NOTE_INPUT_PRESS] =
    gtk_signal_new ("note_input_press", GTK_RUN_FIRST, object_class->type,
		    GTK_SIGNAL_OFFSET (NoteInputClass, note_input_press),
		    note_input_marshal_signal_1, GTK_TYPE_NONE, 1,
		    GTK_TYPE_INT);

  note_input_signals[NOTE_INPUT_RELEASE] =
    gtk_signal_new ("note_input_release", GTK_RUN_FIRST, object_class->type,
		    GTK_SIGNAL_OFFSET (NoteInputClass, note_input_release),
		    note_input_marshal_signal_1, GTK_TYPE_NONE, 1,
		    GTK_TYPE_INT);

  note_input_signals[NOTE_INPUT_TRANSPOSE_CHANGE] =
    gtk_signal_new ("note_input_transpose_change", GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (NoteInputClass, 
				       note_input_transpose_change),
		    note_input_marshal_signal_1, GTK_TYPE_NONE, 1,
		    GTK_TYPE_INT);

  gtk_object_class_add_signals (object_class, note_input_signals,
				LAST_SIGNAL);

  object_class->destroy = note_input_destroy;

  widget_class->key_press_event = note_input_key_press;
  widget_class->key_release_event = note_input_key_release;
}

/* gconstpointer changed to const void * for GTK+ 1.0 compatibility
 */

static gint
compare_gints (const void * a, const void * b)
{
  return (GPOINTER_TO_INT (a) - GPOINTER_TO_INT (b));
}

static void
note_input_init (NoteInput *noteinput)
{
  GTK_WIDGET_SET_FLAGS (noteinput, GTK_CAN_FOCUS);

  noteinput->key_to_func = g_tree_new (compare_gints);
  noteinput->transpose = 36;
  noteinput->keyboard_grab_count = 0;
}

static gint
free_traverse_func (gpointer key, gpointer value, gpointer data)
{
  g_free (value);
  return FALSE;
}

static void
note_input_destroy (GtkObject *object)
{
  NoteInput *noteinput;

  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_NOTE_INPUT (object));

  noteinput = NOTE_INPUT (object);

  g_tree_traverse (noteinput->key_to_func, free_traverse_func,
		   G_IN_ORDER, NULL);
		   
  g_tree_destroy (noteinput->key_to_func);

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

/* This gets called by gdk when a key is pressed.
 */

static gint
note_input_key_press (GtkWidget *widget, GdkEventKey *event)
{
  NoteInput *noteinput = NULL;
  NIFunc *n_i_func = NULL;
  gint old_transpose;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_NOTE_INPUT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  noteinput = NOTE_INPUT (widget);
  old_transpose = noteinput->transpose;

  n_i_func = (NIFunc *)
    g_tree_lookup (noteinput->key_to_func, 
		   GINT_TO_POINTER (event->keyval));
  if (n_i_func)
    {
      switch (n_i_func->type)
	{
	case NI_FUNC_KEY:
	  gtk_signal_emit (GTK_OBJECT (noteinput),
			   note_input_signals [NOTE_INPUT_PRESS], 
			   MIN (127,
				noteinput->transpose + n_i_func->data));
	  break;

	case NI_FUNC_TRANSPOSE_UP:
	  noteinput->transpose ++;
	  break;

	case NI_FUNC_TRANSPOSE_DOWN:
	  noteinput->transpose --;
	  break;

	case NI_FUNC_TRANSPOSE_UP_OCT:
	  noteinput->transpose += 12;
	  break;

	case NI_FUNC_TRANSPOSE_DOWN_OCT:
	  noteinput->transpose -= 12;
	  break;

	case NI_FUNC_OCTAVE_N:
	  noteinput->transpose = (n_i_func->data) * 12;
	  break;

	default:
	  g_assert_not_reached ();
	}

      /* make sure the transpose value is within limits */
      noteinput->transpose = CLAMP (noteinput->transpose, 0, 127);

      if (noteinput->transpose != old_transpose)
	{
	  gtk_signal_emit (GTK_OBJECT (noteinput),
			   note_input_signals [NOTE_INPUT_TRANSPOSE_CHANGE], 
			   noteinput->transpose);
	}

      /* the 0 argument should be `CurrentTime' as defined in <X11/X.h>
       * but then we need to include that file... */
      if (noteinput->keyboard_grab_count == 0)
	{
	  gdk_keyboard_grab (widget->window, TRUE, 0);
	  gdk_key_repeat_disable ();
	}
      noteinput->keyboard_grab_count++;

      return TRUE;
    }

  return FALSE;
}

/* This gets called by gdk when a key is released.
 */

static gint
note_input_key_release (GtkWidget *widget, GdkEventKey *event)
{
  NoteInput *noteinput = NULL;
  NIFunc *n_i_func = NULL;
  
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_NOTE_INPUT (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  noteinput = NOTE_INPUT (widget);

  if (noteinput->keyboard_grab_count > 0)
    {
      noteinput->keyboard_grab_count--;
      if (noteinput->keyboard_grab_count == 0)
	{
	  gdk_keyboard_ungrab (0);
	  gdk_key_repeat_restore ();
	}

      n_i_func = (NIFunc *) 
	g_tree_lookup (noteinput->key_to_func,
		       GINT_TO_POINTER (event->keyval));

      if (n_i_func)
	{
	  if (n_i_func->type == NI_FUNC_KEY)
	    {
	      gtk_signal_emit (GTK_OBJECT (noteinput),
			       note_input_signals [NOTE_INPUT_RELEASE], 
			       MIN (127,
				    noteinput->transpose + n_i_func->data));
	    }
	  /* else, do nothing */

	  return TRUE;
	}
    }

  return FALSE;
}

/* Creates a new NoteInput widget.
 */

GtkWidget *
note_input_new (void)
{
  GtkWidget *noteinput;

  noteinput = gtk_type_new (note_input_get_type ());

  return (noteinput);
}

/* helper for `note_input_read_config'
 */

static void
syntaxerror (gchar *filename, gint lineno)
{
  g_message ("%s: %s: %i: syntax error\n",
	     g_get_prgname (), filename, lineno);
}

/* Helper for note_input_read_config.  Splits a line up in the part
 * before and after an `=' character.  
 * The strings returned in *part1 and *part2 should be g_free'd by the
 * caller if the function returns TRUE.
 * Returns: TRUE on Ok and FALSE if no `=' could be found.
 */

static gboolean
getparts (char *line, char **part1, char **part2)
{
  char *found;

  found = strrchr (line, '=');

  if (found)
    {
      *part1 = g_strndup (line, found-line);
      *part2 = g_strdup (found + 1);
      g_strstrip (*part1);
      g_strstrip (*part2);

      return TRUE;
    }
  else
    return FALSE;
}

/* Helper for note_input_read_config
 * Interprets the `function' on the right hand side of the equality
 * signs, and returns a g_malloc'ed NIFunc, or NULL if it can't
 * find any
 */

NIFunc *
get_n_i_func (char *funcname)
{
  NIFunc *n_i_func = NULL;
  NIFuncType type = NI_FUNC_NONE;
  gint data = 0;

  if (!strcmp (funcname, "TransposeUp"))
    type = NI_FUNC_TRANSPOSE_UP;
  else if (!strcmp (funcname, "TransposeDown"))
    type = NI_FUNC_TRANSPOSE_DOWN;
  else if (!strcmp (funcname, "TransposeUpOctave"))
    type = NI_FUNC_TRANSPOSE_UP_OCT;
  else if (!strcmp (funcname, "TransposeDownOctave"))
    type = NI_FUNC_TRANSPOSE_DOWN_OCT;
  else if (!strncmp (funcname, "Octave", 6) 
	   && (strlen (funcname) > 6)
	   && (isdigit (funcname [6])))
    {
      type = NI_FUNC_OCTAVE_N;
      data = atoi (funcname + 6);
    }
  else
    {
      gint notenum;
      if (notenumber (funcname, &notenum))
	{
	  type = NI_FUNC_KEY;
	  data = notenum;
	}
    }
  
  if (type == NI_FUNC_NONE)
    return NULL;
  else
    {
      n_i_func = g_new (NIFunc, 1);
      n_i_func->type = type;
      n_i_func->data = data;

      return n_i_func;
    }
}

/* Reads in a specified config file.  It's up to the app to choose 
 * what the config file should be called, but mostly: 
 * @sysconfdir@/noteinput or ~/.noteinput
 *
 * Returns: TRUE on success, FALSE on error
 */

gboolean
note_input_read_config (NoteInput *noteinput, gchar *filename)
{
  gboolean err = FALSE;
  gint def_keysym = 0;
  int line_length = 0;
  gint lineno = 0;
  NIFunc *n_i_func;
  char *line = NULL, *l = NULL;
  char *part1 = NULL, *part2 = NULL;
  FILE *file = NULL;

  file = fopen (filename, "r");
  if (file == NULL)
    {
      g_message ("%s: %s: %s\n",
		 g_get_prgname (), filename, g_strerror (errno));
      return FALSE;
    }

  while (1)
    {
      lineno++;

      if (!my_getline (&line, &line_length, file))
	{
	  if (ferror (file))
	    {
	      g_message ("%s: %s: %s\n",
			 g_get_prgname (), filename, g_strerror (errno));
	      err = TRUE;
	    }
	  break;
	}

      l = line;

      if (line_length > 0 && l [0] == '#')
	continue;

      l = g_strstrip (l);

      if (! (*l))
	continue;

      if (! getparts (l, &part1, &part2))
	{
	  syntaxerror (filename, lineno);
	  err = TRUE;
	  break;
	}

      /* interpret the left hand of the equality sign */

      if (*part1 == '\0')
	{
	  syntaxerror (filename, lineno);
	  err = TRUE;
	  break;
	}

      /* keysym string? */

      if (isalpha (*part1))
	{
	  def_keysym = gdk_keyval_from_name (part1);
	  if (def_keysym == 0)
	    {
	      g_message ("%s: %s: %i: undefined keysym `%s'\n",
			 g_get_prgname (), filename, lineno, part1);
	      err = TRUE;
	      break;
	    }
	}

      /* otherwise it must be only one character */

      else if (part1 [1] != '\0')
	{
	  syntaxerror (filename, lineno);
	  err = TRUE;
	  break;
	}

      else
	{
	  def_keysym = part1 [0];
	}

      /* interpret the right hand of the equality sign */

      n_i_func = get_n_i_func (part2);
      if (!n_i_func)
	{
	  syntaxerror (filename, lineno);
	  err = TRUE;
	  break;
	}

      /*      g_print ("`%c' => %i\n", def_char, notenum);
       */
      
      g_tree_insert (noteinput->key_to_func,
		     GINT_TO_POINTER (def_keysym),
		     n_i_func);

      g_free (part1);
      g_free (part2);
      part1 = part2 = NULL;
    }

  /* if the case we jumped out because of error... */
  g_free (part1);
  g_free (part2);

  fclose (file);

  return (!err);
}

/* Return the current transpose base
 */

gint
note_input_get_transpose_base (NoteInput *noteinput)
{
  g_return_val_if_fail (noteinput != NULL, -1);
  g_return_val_if_fail (IS_NOTE_INPUT (noteinput), -1);

  return noteinput->transpose;
}

/* Signal Marshalers */

static void
note_input_marshal_signal_1 (GtkObject *object,
			     GtkSignalFunc func,
			     gpointer func_data,
			     GtkArg *args)
{
  NoteInputSignal1 rfunc;
  
  rfunc = (NoteInputSignal1) func;

  (*rfunc) (object, GTK_VALUE_INT (args[0]), func_data);
}
